/*-
 * Copyright (C) @BABOLO  2002 Dec 17
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 17\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: cmd_select.c,v 1.31 2008/05/26 12:27:06 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

int
pgoblin_perform(pgoblin_exenv *exenv, pgoblin_r *r) {
    void *res;
    int ex = EX_OK;

    ifBLIN_QV4(exenv->options->flags)
        fprintf(stderr, "+pgoblin_perform %s\n", (char*)IO_CTL->text);
    if  (!IO_CTL->text) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_perform: No query");
        return(-1);
    }
    ex = DBASE(R_CONN->flags)->exec(R_CONN, &res, IO_CTL->text, PGOBLIN_CR_OK);
    DBASE(R_CONN->flags)->clear(&res);
    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "-pgoblin_perform %d\n", ex);
    return(ex);
}

int
pgoblin_select(pgoblin_exenv *exenv, pgoblin_r *r) {
    int ex = EX_OK;

    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "+pgoblin_select %d %d %d =%s~\n"
                                             , r[PGO_COUT].r, r[PGO_CCTL].r, r[PGO_CCON].r
                                             , (char*)IO_CTL->text
                                             );
    if  (!IO_CTL->text) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_select: No query");
        ex = EX_NOINPUT;
    } else {
        /** IO_PQ + **/
        ;;  ex = DBASE(R_CONN->flags)->exec(R_CONN, &IO_OUT->pq, IO_CTL->text, PGOBLIN_CR_TUPL);
        ;;  IO_OUT->flagc = R_CONN->flags;
        ;;  IO_OUT->flags |= PGOBLIN_PQRESULT | PGOBLIN_FREEPQRE;
        /** IO_PQ - **/
    }
    ifBLIN_QV4(exenv->options->flags)
        fprintf(stderr, "-pgoblin_select %s\n", IO_OUT->pq ? "OK" : "NULL");
    return(ex);
}

int
pgoblin_display(pgoblin_exenv *exenv, pgoblin_r *r) {
    ssize_t ex;

    ifBLIN_QV4(exenv->options->flags)
        fprintf(stderr, "+pgoblin_display %d %d %d\n", r[PGO_COUT].r, r[PGO_CSTY].r, r[PGO_CIN].r);
    ex = STYLS(R_STYLE->flags)->table(IO_OUT, R_STYLE, IO_IN, 0, -2);
    if  (ex < 0) ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_display: write error");
    ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "-pgoblin_display %d\n", ex);
    return((ex < 0) ? ex : 0);
}

static int
sel(pgoblin_exenv *exenv, pgoblin_r *r, int cond) {
    ssize_t k, e = 0, ex = 0;
    pgoblin_io tr;
    size_t i;
#   define aim  (INT_MAX / 2)
#   define aid                                                                                        \
    ((PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE > (aim / 4)) ? (aim / 4) : (PGOBLIN_BUFMULT * PGOBLIN_BUFSIZE))

    ifBLIN_QV4(exenv->options->flags)
        fprintf( stderr, "+pgoblin_strsel %d %d %d %d =%s~\n"
               , r[PGO_COUT].r, r[PGO_CCTL].r, r[PGO_CCON].r, r[PGO_CSTY].r
               , (char*)IO_CTL->text
               );
    if  (!IO_CTL->text) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_strsel: No query");
        ERROUT(EX_DATAERR, EINVAL);
    }
    tr.flagc = R_CONN->flags;
    if  ((ex = DBASE(R_CONN->flags)->exec(R_CONN, &tr.pq, (char*)IO_CTL->text, cond))) goto out;
    for (k = 0, i = 1; ; ) {
        int typn;

        if  ((ex = DBASE(R_CONN->flags)->getstream(R_CONN, &tr.pq, i, cond))) goto out;
        typn = DBASE(R_CONN->flags)->ntuples(tr.pq);
        e = STYLS(R_STYLE->flags)->table(IO_OUT, R_STYLE, &tr, k, typn ? -1 : -2);
        if  ((e < 0) || (typn <= 0)) break;
        k += typn;
        DBASE(R_CONN->flags)->clear(&tr.pq);
        ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "e=%d=>%d aim=%d i %d => ", typn, e, aim, i);
        if  (e < (aim * 2)) {
            if  (i < (aid / 2)) i = i * 2; else i = aid;
        } else if (e < aim) {
            if  (i < (aid / aim * e)) i = i / e * aim; else i = aid;
        } else if ((e / 2) > aim) {
            i = i / 4;
        } else if (e > aim) {
            i = i / e * aim;
        }
        ifBLIN_QV4(exenv->options->flags) fprintf(stderr, "%d\n", i);
    }
    DBASE(R_CONN->flags)->clear(&tr.pq);
    if  (e < 0) ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_strsel: write error");
out:
/* END      ? */
    ex = DBASE(R_CONN->flags)->endstream(R_CONN, cond);
    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "-pgoblin_strsel %d\n", 0);
    return((e < 0) ? e : ex);
}

int
pgoblin_strsel(pgoblin_exenv *exenv, pgoblin_r *r) {
    return(sel(exenv, r, PGOBLIN_CR_SSEL));
}

int
pgoblin_binsel(pgoblin_exenv *exenv, pgoblin_r *r) {
    return(sel(exenv, r, PGOBLIN_CR_BSEL));
}

int
pgoblin_binpre(pgoblin_exenv *exenv, pgoblin_r *r) {
    void *res;
    int ex = EX_OK;
    char *q;

    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "+pgoblin_binpre %d %d %d =%s~\n"
                                      , r[PGO_COUT].r, r[PGO_CCTL].r, r[PGO_CCON].r
                                      , (char*)IO_CTL->text
                                      );
    if  (!IO_CTL->text) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_binpre: No query");
        ERROUT(EX_DATAERR, EINVAL);
    }
    if  (!R_CONN->intran && (ex = DBASE(R_CONN->flags)->exec(R_CONN, &res, "BEGIN", PGOBLIN_CR_OK)))
        goto out;
    asprintf(&q, "DECLARE pgoblin BINARY NO SCROLL CURSOR FOR %s", (char*)IO_CTL->text);
    if  (q == NULL) {
        ifBLIN_QV1(exenv->options->flags) warnx("pgoblin_binpre: No memory");
        ERROUT(EX_UNAVAILABLE, ENOMEM);
    }
    ex = DBASE(R_CONN->flags)->exec(R_CONN, &res, q, PGOBLIN_CR_OK);
    DBASE(R_CONN->flags)->clear(&res);
    free(q);
    if  (ex) goto out;
    /** IO_PQ + **/
    ;;  ex = DBASE(R_CONN->flags)->exec(R_CONN, &IO_OUT->pq, "FETCH 2147483647 FROM pgoblin", PGOBLIN_CR_TUPL);
    ;;  IO_OUT->flagc = R_CONN->flags;
    ;;  IO_OUT->flags |= PGOBLIN_PQRESULT | PGOBLIN_FREEPQRE;
    /** IO_PQ - **/
    ex = DBASE(R_CONN->flags)->exec(R_CONN, &res, "CLOSE pgoblin", PGOBLIN_CR_OK);
    DBASE(R_CONN->flags)->clear(&res);
out:
    if  (!R_CONN->intran) ex = DBASE(R_CONN->flags)->exec(R_CONN, &res, "END", PGOBLIN_CR_OK);
    DBASE(R_CONN->flags)->clear(&res);
    ifBLIN_QV4(exenv->options->flags) fprintf( stderr, "-pgoblin_binpre %d\n", 0);
    return(ex);
}
